#!/usr/bin/env python

"""
Convert a graph between NetworkX and OrientDB
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import copy
import json
import time

import numpy as np
import networkx as nx
import pyorient.otypes
from pyorient.utils import get_hash

from .utils import _find_field_types
from ..utils import byteify, chunks

def as_nx(nodes=[], edges=[], force_rid=False, deepcopy = True):
    """
    Converts OrientDB Gremlin query results into a NetworkX MultiDiGraph.

    Parameters
    ----------
    nodes : list of pyorient.otypes.OrientRecord
        OrientDB node query results.
    edges : list of pyorient.otypes.OrientRecord
        OrientDB edge query results.
    force_rid : bool
        If True, always use the OrientDB RID as the node identifier in the
        returned graph. Otherwise, use 'id' property as the
        node identifier if it is present.

    Results
    -------
    g : networkx.MultiDiGraph
        Constructed multigraph containing query results. The OrientDB class of each
        node and edge is stored in the 'class' attribute
        of the corresponding nodes and edges in the result `g`.
    """

    # XXX what should happen if a node/edge in OrientDB contains a 'class' attribute?
    g = nx.MultiDiGraph()
    rid_to_id = {}
    for i, node in enumerate(nodes):
        # Don't let function alter the original records:
        if deepcopy:
            tmp = copy.deepcopy(node.oRecordData)
        else:
            tmp = node.oRecordData
        props = {}
        for k, v in tmp.items():
            if isinstance(v, pyorient.otypes.OrientBinaryObject):
                continue
            if isinstance(k, str) and k.startswith('_'):
                continue
            if isinstance(v, pyorient.otypes.OrientRecordLink):
                props[k] = get_hash(v)
            elif (isinstance(v,list) and v and
                       isinstance(v[0], pyorient.otypes.OrientRecordLink)):
                props[k] = [get_hash(n) for n in v]
            else:
                props[k] = v
        # props_keys = list(props.keys())
        # for k in props_keys:
        #
        #     # Discard binary objects:
        #     if isinstance(props[k], pyorient.otypes.OrientBinaryObject):
        #         del props[k]
        #
        #     # Replace record links with their corresponding RIDs:
        #     #elif isinstance(props[k], pyorient.otypes.OrientRecordLink):
        #     #    props[k] = props[k].get_hash()
        #
        #     # Remove record links
        #     elif isinstance(props[k], pyorient.otypes.OrientRecordLink):
        #         del props[k]
        #
        #     # Remove list of links
        #     elif (isinstance(props[k],list) and props[k] and
        #           isinstance(props[k][0], pyorient.otypes.OrientRecordLink)):
        #         del props[k]
        #
        #     # Remove properties whose name is a string that starts with '_'; they
        #     # are for special OrientDB purposes:
        #     elif isinstance(k, str) and k.startswith('_'):
        #         del props[k]

        # Save the OrientDB class:
        props['class'] = node._class

        # If the node has an 'id' attribute, use that as the NetworkX node
        # identifier:
        if 'id' in props and not force_rid:
            id = props['id']
            del props['id']
        else:
            id = props.get('rid', node._rid)

        g.add_node(id, **props)

        rid_to_id[props.get('rid', node._rid)] = id

    for edge in edges:
        # Don't let function alter the original records:
        if deepcopy:
            tmp = copy.deepcopy(edge.oRecordData)
        else:
            tmp = edge.oRecordData
        in_rid = tmp['in'].get_hash()
        out_rid = tmp['out'].get_hash()
        # del props['in']
        # del props['out']
        props = {k: v for k, v in tmp.items() if k not in ['in', 'out']}

        # Save the OrientDB class:
        props['class'] = edge._class
        g.add_edge(rid_to_id[out_rid], rid_to_id[in_rid],
                   **props)
    return g

def orient_to_nx(client, node_query='', edge_query='', force_rid=False):
    """
    Query OrientDB and return results as a NetworkX MultiDiGraph.

    Parameters
    ----------
    client : pyorient.orient.OrientDB
        OrientDB interface.
    node_query : str
        Gremlin query that returns a collection of nodes.
    edge_query : str
        Gremlin query that returns a collection of edges.
    force_rid : bool
        If True, always use the OrientDB RID as the node identifier in the
        returned graph. Otherwise, use 'id' property as the
        node identifier if it is present.

    Results
    -------
    g : networkx.MultiDiGraph
        Constructed multigraph containing query results. The OrientDB class of each
        node and edge is stored in the 'class' attribute
        of the corresponding nodes and edges in the result `g`.
    """

    if node_query:
        nodes = client.gremlin(node_query)
    else:
        nodes = []
    if edge_query:
        edges = client.gremlin(edge_query)
    else:
        edges = []
    return as_nx(nodes, edges, force_rid)

def nx_to_orient(client, g):
    """
    Converts NetworkX MultiDiGraph to OrientDB graph.

    Parameters
    ----------
    client : pyorient.orient.OrientDB
        OrientDB interface.
    g : networkx.MultiDiGraph
        Graph to convert to OrientDB.

    Notes
    -----
    The 'class' attribute of each node and edge in `g` is assumed to be the
    OrientDB class name to use when creating the corresponding nodes and edges
    in the database. If no 'class' attribute is specified, the node and edge
    class names are assumed to be 'V' and 'E', respectively.

    Node IDs are discarded upon creation of the new graph.
    """

    assert isinstance(g, (nx.DiGraph, nx.MultiDiGraph))

    # This assumes that each OrientDB class has a single cluster:
    N = 10
    id_to_rid = {}
    for chunk in chunks(g.nodes(data=True), N):
        cmd_list = []
        id_list = []
        for i, (id, props) in enumerate(chunk):
            # Remove class name from properties inserted into database (but don't
            # clobber the input graph):
            props = copy.deepcopy(props)
            if props.has_key('class'):
                cls = props['class']
                del props['class']
            else:
                cls = 'V'

            # Save original node ID:
            assert 'id' not in props
            props['id'] = id

            # Add @fieldTypes field to force proper storage of types:
            ft = _find_field_types(props)
            if ft:
                props['@fieldTypes'] = ft

            id_list.append(id)
            cmd_list.append('let $a%s = create vertex %s content %s;' % \
                            (i, cls, json.dumps(byteify(props))))
        cmd = 'begin;'+''.join(cmd_list)+'commit;'+\
              ('return [%s];' % ','.join(['$a%s' % i for i in range(len(chunk))]))
        rec_list = client.batch(cmd)
        for id, r in zip(id_list, rec_list):
            id_to_rid[id] = r._rid

    for chunk in chunks(g.edges(data=True), N):
        cmd_list = []
        for from_id, to_id, props in chunk:
            # Remove class name from properties inserted into database (but don't
            # clobber the input graph):
            props = copy.deepcopy(props)
            if props.has_key('class'):
                cls = props['class']
                del props['class']
            else:
                cls = 'E'

            # Add @fieldTypes field to force proper storage of types:
            ft = _find_field_types(props)
            if ft:
                props['@fieldTypes'] = ft

            cmd_list.append('create edge %s from %s to %s content %s;' % \
                (cls, id_to_rid[from_id], id_to_rid[to_id],
                 json.dumps(byteify(props))))
        cmd = 'begin;'+''.join(cmd_list)+'commit;'
        client.batch(cmd)



#!/usr/bin/env python

"""
Neuroarch data models.
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import inspect

from pyorient.ogm.declarative import declarative_node, \
    declarative_relationship
from pyorient.ogm.property import Property, EmbeddedMap, EmbeddedSet, String, EmbeddedList, Boolean, Integer, Double

#import neuroarch.conv.pd as pd
#import neuroarch.conv.nx as nx
import neuroarch.utils as utils

from .query import QueryWrapper, QueryString

def create_efficiently(graph, registry):
    """
    Efficiently create classes in OrientDB database.

    Runs `graph.create_all(registry)` if any of the classes in the registry are
    absent; otherwise, runs the faster method `graph.include(registry)`.

    Parameters
    ----------
    graph : pyorient.ogm.graph.Graph
        OrientDB graph.
    registry : collections.OrderedDict
        Registry of pyorient OGM classes for which to create OrientDB classes.
    """

    cluster_ids = utils.get_cluster_ids(graph.client)
    if not set(registry.keys()).issubset(cluster_ids.keys()):
        graph.create_all(registry)
    else:
        graph.include(registry)

class NeuroarchNodeMixin(object):
    def gremlin(self, script, args=None, namespace=None):
        """
        Return results of Gremlin query as mapped objects.
        """

        return self._graph.gremlin(script, args, namespace)

    def gremlin_raw(self, script, args=None, namespace=None):
        """
        Return results of Gremlin query as pyorient records.
        """

        graph = self._graph
        script_body = graph.scripts.script_body(script, args, namespace)
        if script_body:
            response = graph.client.gremlin(script_body)
        else:
            response = graph.client.gremlin(script)
        return response

    def update(self, **props):
        """
        Update record in database with specified properties.
        """

        self._graph.client.command('update %s merge %s where @rid = %s' % \
                (self.element_type, props, self._id))

    def get_props(self):
        """
        Retrieve record properties from database.
        """

        props = self._graph.client.query('select * from %s' % self._id)[0]
        return utils.orientrecord_to_dict(props)
    
    @property
    def props(self):
        """
        Returns record properties that have been retrieved.
        """
        return {k: getattr(self, k) for k, v in inspect.getmembers(type(self)) if isinstance(v, Property)}

    # XXX add option to these methods to control whether both nodes and edges
    # are returned:
    def owns(self, levels=1, **kwargs):
        """
        Retrieve nodes/edges owned by this node at a specific level of the ownership hierarchy.
        """

        q = QueryWrapper(self._graph, QueryString('select from %s' % self._id, 'sql'))
        return q.owns(levels=levels, **kwargs)

    def traverse_owns(self, **kwargs):
        """
        Traverse nodes/edges owned by this node at a specific level of the ownership hierarchy.
        """

        q = QueryWrapper(self._graph, QueryString('select from %s' % self._id, 'sql'))
        return q.traverse_owns(**kwargs)

    def owned_by(self, levels=1, **kwargs):
        """
        Retrieve nodes/edges that own this node at a specific level of the ownership hierarchy.
        """

        q = QueryWrapper(self._graph, QueryString('select from %s' % self._id, 'sql'))
        return q.owned_by(levels=levels, **kwargs)

    def traverse_owned_by(self, **kwargs):
        """
        Traverse nodes/edges that own this node at a specific level of the ownership hierarchy.
        """

        q = QueryWrapper(self._graph, QueryString('select from %s' % self._id, 'sql'))
        return q.traverse_owned_by(**kwargs)

class Node(NeuroarchNodeMixin, declarative_node()):
    element_type = 'Node'
    element_plural = 'Nodes'

class Relationship(declarative_relationship()):
    label = 'Relationship'

class MetaData(Node):
    element_type = 'MetaData'
    element_plural = 'MetaDatas'
    created_by = EmbeddedMap(linked_to=String(), nullable=False, unique=False)
    version = String(nullable=False, unique=False, indexed=True)
    # NeuroArch_version = String(nullable=False, unique=False, indexed=True)
    # min_NeuroArch_version_supported = String(nullable=False, unique=False, indexed=True)
    # OrientDB_version = String(nullable=False, unique=False, indexed=True)
    maintainer = EmbeddedMap(linked_to=String(), nullable=False, unique=False)
    
class Species(Node):
    element_type = 'Species'
    element_plural = 'Species'
    name = String(nullable=False, unique=False, indexed=True)
    stage = String(nullable=False, unique=False, indexed=True)
    sex = String(nullable=False, unique=False, indexed=True)
    synonyms = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)

# Biological data nodes:
class BioNode(Node):
    element_type = 'BioNode'
    element_plural = 'BioNodes'

class DataSource(BioNode):
    element_type = 'DataSource'
    element_plural = 'DataSources'
    name = String(nullable=False, unique=False, indexed=True)
    version = String(nullable=False, unique=False, indexed=True)
    description = String(nullable=True, unique=False, indexed=False)
    url = String(nullable=True, unique=False, indexed=False)

class Subsystem(BioNode):
    element_type = 'Subsystem'
    element_plural = 'Subsystems'
    name = String(nullable=False, unique=False, indexed=True)
    synonyms = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)

class Neuropil(BioNode):
    element_type = 'Neuropil'
    element_plural = 'Neuropils'
    name = String(nullable=False, unique=False, indexed=True)
    synonyms = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)

class Subregion(BioNode):
    element_type = 'Subregion'
    element_plural = 'Subregions'
    name = String(nullable=False, unique=False, indexed=True)
    synonyms = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)

class Tract(BioNode):
    element_type = 'Tract'
    element_plural = 'Tracts'
    name = String(nullable=False, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)
    neuropils = EmbeddedSet(linked_to=String(), nullable=True, unique=False, indexed=True)

class BioSensor(BioNode):
    element_type = 'BioSensor'
    element_plural = 'BioSensors'
    name = String(nullable=False, unique=False, indexed=True)

class Circuit(BioNode):
    element_type = 'Circuit'
    element_plural = 'Circuits'
    name = String(nullable=False, unique=False, indexed=True)

class Ommatidium(Circuit):
    element_type = 'Ommatidium'
    element_plural = 'Ommatidia'
    name = String(nullable=False, unique=False, indexed=True)

class Cartridge(Circuit):
    element_type = 'Cartridge'
    element_plural = 'Cartridges'
    name = String(nullable=False, unique=False, indexed=True)

class Glomerulus(Circuit):
    element_type = 'Glomerulus'
    element_plural = 'Glomeruli'
    name = String(nullable=False, unique=False, indexed=True)

class Column(Circuit):
    element_type = 'Column'
    element_plural = 'Columns'
    name = String(nullable=False, unique=False, indexed=True)

class NeuronAndFragment(BioNode):
    element_type = 'NeuronAndFragment'
    element_plural = 'NeuronAndFragments'

class Neuron(NeuronAndFragment):
    element_type = 'Neuron'
    element_plural = 'Neurons'
    name = String(nullable=False, unique=False, indexed=True)
    locality = Boolean(nullable=True, unique=False, indexed=True)
    label = String(nullable=True, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    synonyms = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)
    referenceId = String(nullable=True, unique=False, indexed=True)
    info = EmbeddedMap(nullable=True, unique=False, indexed=True)

class NeuronFragment(NeuronAndFragment):
    element_type = 'NeuronFragment'
    element_plural = 'NeuronFragments'
    name = String(nullable=False, unique=False, indexed=True)
    label = String(nullable=True, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    referenceId = String(nullable=True, unique=False, indexed=True)
    info = EmbeddedMap(nullable=True, unique=False, indexed=True)

class NeuronTerminal(BioNode):
    element_type = 'NeuronTerminal'
    element_plural = 'NeuronTerminals'
    name = String(nullable=False, unique=False, indexed=True)
    synonyms = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)

class Synapse(BioNode):
    element_type = 'Synapse'
    element_plural = 'Synapses'
    name = String(nullable=False, unique=False, indexed=True)
    N = Integer(nullable=True, unique=False, indexed=True)
    NHP = Integer(nullable = True, unique=False, indexed = True)
    uname = String(nullable=True, unique=False, indexed=True)

class InferredSynapse(BioNode):
    element_type = 'InferredSynapse'
    element_plural = 'InferredSynapses'
    name = String(nullable=False, unique=False, indexed=True)
    N = Integer(nullable=True, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)

class GapJunction(BioNode):
    element_type = 'GapJunction'
    element_plural = 'GapJunctions'
    name = String(nullable=False, unique=False, indexed=True)

class PhotoreceptorCell(Neuron):
    element_type = 'PhotoreceptorCell'
    element_plural = 'PhotoreceptorCells'
    #name = String(nullable=False, unique=False, indexed=True)

class ArborizationData(BioNode):
    element_type = 'ArborizationData'
    element_plural = 'ArborizationDatas'
    dendrites = EmbeddedMap(linked_to=Integer(), nullable=True, unique=False, indexed=True)
    axons = EmbeddedMap(linked_to=Integer(), nullable=True, unique=False, indexed=True)
    synapses = EmbeddedMap(linked_to=Integer(), nullable=True, unique=False, indexed=True)
    name = String(nullable=True, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    type = String(nullable=True, unique=False, indexed=True)

class GeneticData(BioNode):
    element_type = 'GeneticData'
    element_plural = 'GeneticDatas'
    name = String(nullable=False, unique=False, indexed=True)

class MorphologyData(BioNode):
    element_type = 'MorphologyData'
    element_plural = 'MorphologyDatas'
    name = String(nullable=False, unique=False, indexed=True)
    morph_type = String(nullable=True, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    x = EmbeddedList(linked_to=Double(), nullable=True, unique=False, indexed=False)
    y = EmbeddedList(linked_to=Double(), nullable=True, unique=False, indexed=False)
    z = EmbeddedList(linked_to=Double(), nullable=True, unique=False, indexed=False)
    r = EmbeddedList(linked_to=Double(), nullable=True, unique=False, indexed=False)
    parent = EmbeddedList(linked_to=Integer(), nullable=True, unique=False, indexed=False)
    identifier = EmbeddedList(linked_to=Integer(), nullable=True, unique=False, indexed=False)
    sample = EmbeddedList(linked_to=Integer(), nullable=True, unique=False, indexed=False)
    confidence = EmbeddedList(linked_to=Double(), nullable=True, unique=False, indexed=False)
    vertices = EmbeddedList(linked_to=Double(), nullable=True, unique=False, indexed=False)
    faces = EmbeddedList(linked_to=Integer(), nullable=True, unique=False, indexed=False)

class NeurotransmitterData(BioNode):
    element_type = 'NeurotransmitterData'
    element_plural = 'NeurotransmitterDatas'
    name = String(nullable=False, unique=False, indexed=True)
    Transmitters = EmbeddedList(linked_to=String(), nullable=False, unique=False, indexed=True)

# Circuit design nodes:
class DesignNode(Node):
    element_type = 'DesignNode'
    element_plural = 'DesignNodes'

class Version(DesignNode):
    element_type = 'Version'
    element_plural = 'Versions'
    version = String(nullable=False, unique=False, indexed=True)

class ExecutableCircuit(DesignNode):
    element_type = 'ExecutableCircuit'
    element_plural = 'ExecutableCircuits'
    name = String(nullable=False, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)

class CircuitDiagram(DesignNode):
    element_type = 'CircuitDiagram'
    element_plural = 'CircuitDiagrams'
    name = String(nullable=False, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)
    diagrams = EmbeddedMap(linked_to=String(), nullable=False, unique=False, indexed=False)
    submodules = EmbeddedMap(linked_to=String(), nullable=True, unique=False, indexed=False)

class LPU(DesignNode):
    element_type = 'LPU'
    element_plural = 'LPUs'
    name = String(nullable=False, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)

class Interface(DesignNode):
    element_type = 'Interface'
    element_plural = 'Interfaces'
    name = String(nullable=False, unique=False, indexed=True)

class Pattern(DesignNode):
    element_type = 'Pattern'
    element_plural = 'Patterns'
    name = String(nullable=False, unique=False, indexed=True)
    version = String(nullable=True, unique=False, indexed=True)

class Port(DesignNode):
    element_type = 'Port'
    element_plural = 'Ports'
    selector = String(nullable=True, unique=False, indexed=True)
    port_type = String(nullable=True, unique=False)
    port_io = String(nullable=True, unique=False)

class CircuitModel(DesignNode):
    element_type = 'CircuitModel'
    element_plural = 'CircuitModels'
    name = String(nullable=False, unique=False, indexed=True)

class OmmatidiumModel(CircuitModel):
    element_type = 'OmmatidiumModel'
    element_plural = 'OmmatidiumModels'
    #name = String(nullable=False, unique=False, indexed=True)

class CartridgeModel(CircuitModel):
    element_type = 'CartridgeModel'
    element_plural = 'CartridgeModels'
    #name = String(nullable=False, unique=False, indexed=True)

class CRModel(CircuitModel):
    element_type = 'CRModel'
    element_plural = 'CRModels'
    #name = String(nullable=False, unique=False, indexed=True)

class ColumnModel(CircuitModel):
    element_type = 'ColumnModel'
    element_plural = 'ColumnModels'
    #name = String(nullable=False, unique=False, indexed=True)

class NeuronModel(DesignNode):
    element_type = 'NeuronModel'
    element_plural = 'NeuronModels'
    name = String(nullable=False, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    params = EmbeddedMap(nullable=True, unique=False, indexed=True)
    states = EmbeddedMap(nullable=True, unique=False, indexed=True)

class MembraneModel(NeuronModel):
    element_type = 'MembraneModel'
    element_plural = 'MembraneModels'

class AxonHillockModel(NeuronModel):
    element_type = 'AxonHillockModel'
    element_plural = 'AxonHillockModels'

class DendriteModel(DesignNode):
    element_type = 'DendriteModel'
    element_plural = 'DendriteModels'
    name = String(nullable=False, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    params = EmbeddedMap(nullable=True, unique=False, indexed=True)
    states = EmbeddedMap(nullable=True, unique=False, indexed=True)

class PhotoreceptorModel(MembraneModel):
    element_type = 'PhotoreceptorModel'
    element_plural = 'PhotoreceptorModels'
    spiking = False
    # num_microvilli = Integer(nullable=True, unique=False, indexed=True)

# Added for AdaptiveNarx
class NarxAdaptive(MembraneModel):
    element_type = 'NarxAdaptive'
    element_plural = 'NarxAdaptives'
    #name = String(nullable=False, unique=False, indexed=True)

class MorrisLecar(MembraneModel):
    element_type = 'MorrisLecar'
    element_plural = 'MorrisLecars'
    spiking = False
    # V1 = Double(nullable=True, unique=False, indexed=True)
    # V2 = Double(nullable=True, unique=False, indexed=True)
    # V3 = Double(nullable=True, unique=False, indexed=True)
    # V4 = Double(nullable=True, unique=False, indexed=True)
    # phi = Double(nullable=True, unique=False, indexed=True)
    # offset = Double(nullable=True, unique=False, indexed=True)
    # V_L = Double(nullable=True, unique=False, indexed=True)
    # V_Ca = Double(nullable=True, unique=False, indexed=True)
    # V_K = Double(nullable=True, unique=False, indexed=True)
    # g_L = Double(nullable=True, unique=False, indexed=True)
    # g_Ca = Double(nullable=True, unique=False, indexed=True)
    # g_K = Double(nullable=True, unique=False, indexed=True)
    # initV = Double(nullable=True, unique=False, indexed=True)
    # initn = Double(nullable=True, unique=False, indexed=True)

class HodgkinHuxley(AxonHillockModel):
    element_type = 'HodgkinHuxley'
    element_plural = 'HodgkinHuxleys'
    spiking = True
    # initV = Double(nullable=True, unique=False, indexed=True)
    # initn =  Double(nullable=True, unique=False, indexed=True)
    # initm = Double(nullable=True, unique=False, indexed=True)
    # inith = Double(nullable=True, unique=False, indexed=True)
    # g_K = Double(nullable=True, unique=False, indexed=True)
    # g_Na = Double(nullable=True, unique=False, indexed=True)
    # g_l =  Double(nullable=True, unique=False, indexed=True)

class HodgkinHuxleyFull(AxonHillockModel):
    element_type = 'HodgkinHuxleyFull'
    element_plural = 'HodgkinHuxleyFulls'
    spiking = True
    # initV = Double(nullable=True, unique=False, indexed=True)
    # initn =  Double(nullable=True, unique=False, indexed=True)
    # initm = Double(nullable=True, unique=False, indexed=True)
    # inith = Double(nullable=True, unique=False, indexed=True)
    # g_K = Double(nullable=True, unique=False, indexed=True)
    # g_Na = Double(nullable=True, unique=False, indexed=True)
    # g_L =  Double(nullable=True, unique=False, indexed=True)
    # E_K =Double(nullable=True, unique=False, indexed=True)
    # E_Na = Double(nullable=True, unique=False, indexed=True)
    # E_L = Double(nullable=True, unique=False, indexed=True)

class LeakyIAF(AxonHillockModel):
    element_type = 'LeakyIAF'
    element_plural = 'LeakyIAFs'
    spiking = True
    # initV = Double(nullable=True, unique=False, indexed=True)
    # threshold = Double(nullable=True, unique=False, indexed=True)
    # reset_potential = Double(nullable=True, unique=False, indexed=True)
    # capacitance = Double(nullable=True, unique=False, indexed=True)
    # resting_potential = Double(nullable=True, unique=False, indexed=True)
    # resistance = Double(nullable=True, unique=False, indexed=True)

class LeakyIAFwithRefractoryPeriod(AxonHillockModel):
    element_type = 'LeakyIAFwithRefractoryPeriod'
    element_plural = 'LeakyIAFwithRefractoryPeriods'
    spiking = True
    # initV = Double(nullable=True, unique=False, indexed=True)
    # threshold = Double(nullable=True, unique=False, indexed=True)
    # reset_potential = Double(nullable=True, unique=False, indexed=True)
    # capacitance = Double(nullable=True, unique=False, indexed=True)
    # resting_potential = Double(nullable=True, unique=False, indexed=True)
    # time_constant = Double(nullable=True, unique=False, indexed=True)
    # refractory_period = Double(nullable=True, unique=False, indexed=True)
    # bias_current = Double(nullable=True, unique=False, indexed=True)

class BufferVoltage(MembraneModel):
    element_type = 'BufferVoltage'
    element_plural = 'BufferVoltages'

class BufferPhoton(MembraneModel):
    element_type = 'BufferPhoton'
    element_plural = 'Bufferphotons'

class Aggregator(DendriteModel):
    element_type = 'Aggregator'
    element_plural = 'Aggregators'

class SynapseModel(DesignNode):
    element_type = 'SynapseModel'
    element_plural = 'SynapseModels'
    name = String(nullable=False, unique=False, indexed=True)
    uname = String(nullable=True, unique=False, indexed=True)
    params = EmbeddedMap(nullable=True, unique=False, indexed=True)
    states = EmbeddedMap(nullable=True, unique=False, indexed=True)
    # reverse = Double(nullable=True, unique=False, indexed=True)
    # gmax = Double(nullable=True, unique=False, indexed=True)

class AlphaSynapse(SynapseModel):
    element_type = 'AlphaSynapse'
    element_plural = 'AlphaSynapses'
    link_pre = 'spike_state'
    link_post = None
    # ar = Double(nullable=True, unique=False, indexed=True)
    # ad = Double(nullable=True, unique=False, indexed=True)

class PowerGPotGPot(SynapseModel):
    element_type = 'PowerGPotGPot'
    element_plural = 'PowerGPotGPots'
    link_pre = 'gpot'
    link_post = None

class SynapseAMPA(SynapseModel):
    element_type = 'SynapseAMPA'
    element_plural = 'SynapseAMPAs'
    link_pre = 'spike_state'
    link_post = None
    # st = Double(nullable=True, unique=False, indexed=True)

class SynapseGABA(SynapseModel):
    element_type = 'SynapseGABA'
    element_plural = 'SynapseGABAs'
    link_pre = 'spike_state'
    link_post = None
    # st = Double(nullable=True, unique=False, indexed=True)

class SynapseNMDA(SynapseModel):
    element_type = 'SynapseNMDA'
    element_plural = 'SynapseNMDAs'
    link_pre = 'spike_state'
    link_post = 'V'
    # st = Double(nullable=True, unique=False, indexed=True)
    # Mg = Double(nullable=True, unique=False, indexed=True)

class SigmoidSynapse(SynapseModel):
    element_type = 'SigmoidSynapse'
    element_plural = 'SigmoidSynapses'
    link_pre = 'V'
    link_post = None
    # threshold = Double(nullable=True, unique=False, indexed=True)
    # scale = Double(nullable=True, unique=False, indexed=True)
    # slope = Double(nullable=True, unique=False, indexed=True)

class Owns(Relationship):
    label = 'Owns'

class SendsTo(Relationship):
    label = 'SendsTo'
    variable = String(nullable=True, unique=False, indexed=True)

class HasData(Relationship):
    label = 'HasData'

class Requires(Relationship):
    label = 'Requires'

class ArborizesIn(Relationship):
    label = 'ArborizesIn'
    kind = EmbeddedSet(linked_to=String(), nullable=False, unique=False, indexed=True)
    N_axons = Integer(nullable=True, unique=False, indexed=True)
    N_dendrites = Integer(nullable=True, unique=False, indexed=True)

class QueryResult(Node):
    element_type = 'QueryResult'
    element_plural = 'QueryResults'
    tag =  String(nullable=False, unique=True, indexed=True)
    keywords = EmbeddedList(linked_to=String(), nullable=True, unique=False, indexed=True)
    FFBOdata = EmbeddedMap(linked_to=String(), nullable=True, unique=False, indexed=True)

class QueryOwns(Relationship):
    label = 'QueryOwns'

class Models(Relationship):
    label = 'Models'
    version = String(nullable=True, unique=False, indexed=True)

class HasQueryResults(Relationship):
    label = 'HasQueryResults'

Data_Types = ['GeneticData', 'NeurotransmitterData', 'MorphologyData', 'ArborizationData']



#!/usr/bin/env python

"""
"""

import copy

import networkx as nx

import neuroarch.query as query

def to_nk(graph, node_query):
    q = neuroarch.query.QueryWrapper(graph, node_query)

    # Extract nodes and edges as networkx.MultiDiGraph:
    g = q.get_as('nx')

    # Convert to Neurokernel data structures:
    return na_to_nk(g)

def na_to_nk(g):
    """
    Convert graph extracted from Neuroarch to Neurokernel-compatible data structures.

    Parameters
    ----------
    g : networkx.MultiDiGraph

    Returns
    ------
    n_dict : dict of dict of list
        Each key of `n_dict` is the name of a neuron model; the values
        are dicts that map each attribute name to a list that contains the
        attribute values for each neuron.
    s_dict : dict of dict of list
        Each key of `s_dict` is the name of a synapse model; the values are
        dicts that map each attribute name to a list that contains the
        attribute values for each each neuron.
    """

    assert isinstance(g, nx.MultiDiGraph)
    nodes = g.nodes()

    # Lexicographically sort neurons by RID so that consecutive neurons
    # of the same type are the same in Neurokernel:
    neurons = [(rid, g.node[rid]) for rid in nodes \
               if g.node[rid]['class'] == 'neuron']
    neurons.sort()

    # Output neuron data structure:
    n_dict = {}

    # Maps RIDs to consecutive indices in model param lists:
    rid_id_map = {}

    # Maps RIDs of nodes of a particular model to consecutive indices:
    rid_model_id_map = {}
    for id, (rid, neu) in enumerate(neurons):
        neu = copy.deepcopy(neu)
        model = neu['model']

        # Ensure a selector is defined for output ports:
        if 'public' not in neu:
            neu['public'] = False
        if 'selector' not in neu:
            neu['selector'] = ''

        # If a neuron model has not appeared before, add it to n_dict:
        if model not in n_dict:
            n_dict[model] = {k: [] for k in neu.keys() + ['id']}
            rid_model_id_map[model] = {}

        # Add neuron data to n_dict subdictionary:
        for key in neu.keys():
            n_dict[model][key].append(neu[key])

        # Neurons of the same model should have the same attributes:
        assert(set(n_dict[model].keys()) == set(neu.keys() + ['id']))

        # Save map between RID and ID of current node:
        rid_id_map[rid] = id
        rid_model_id_map[model][rid] = len(n_dict[model]['id'])
            
        n_dict[model]['id'].append(int(id))

    # Process output ports, i.e., those with class 'port' that have an incoming
    # edge from a node with class 'neuron' (XXX should check that edges are of
    # class 'data'):
    ports = [rid for rid in nodes if g.node[rid]['class'] == 'port']
    out_ports = [(port_rid, neuron_rid) for port_rid in ports \
                 for neuron_rid in g.predecessors(port_rid) \
                 if g.node[neuron_rid]['class'] == 'neuron']
    for port_rid, neuron_rid in out_ports:
        sel = g.node[port_rid]['selector']
        model = g.node[neuron_rid]['model']
        n_dict[model]['selector'][rid_model_id_map[model][neuron_rid]] = sel

    # Process input ports, i.e., those with class 'port' that have an outgoing
    # edge to a node with class 'synapse' (XXX should check that edges are of
    # class 'data'):
    in_ports = [(port_rid, g.node[port_rid]) for port_rid in ports \
                if 'synapse' in \
                [g.node[synapse_rid]['class'] for synapse_rid \
                 in g.successors(port_rid)]]
    for id, (port_rid, port_data) in enumerate(in_ports):
        port_data = copy.deepcopy(port_data)
        model = port_data['model']
        assert 'selector' in port_data
        if model == 'port_in_gpot':
            port_data['spiking'] = False
            port_data['public'] = False
        else:
            port_data['spiking'] = True
            port_data['public'] = False

        if 'public' not in port_data:
            port_data['public'] = False
        if model not in n_dict:
            n_dict[model] = {k: [] for k in port_data.keys() + ['id']}
            rid_model_id_map[model] = {}
            
        assert set(n_dict[model].keys()) == set(port_data.keys()+['id'])

        for key in port_data:
            n_dict[model][key].append(port_data[key])

        rid_id_map[rid] = id
        rid_model_id_map[model][rid] = len(n_dict[model]['id'])
        n_dict[model]['id'].append(int(id))

    # Remove duplicate neuron model info:
    for val in n_dict.values():
        val.pop('model')
    if not n_dict:
        n_dict = dict()

    # Process synapses:
    synapses = [(rid_id_map[pre_rid], rid_id_map[post_rid], g.node[synapse_rid]) \
                for synapse_rid in nodes \
                if g.node[synapse_rid]['class'] == 'synapse' \
                for pre_rid in g.predecessors(synapse_rid) \
                if g.node[pre_rid]['class'] == 'neuron' \
                for post_rid in g.successors(synapse_rid) \
                if g.node[post_rid]['class'] == 'neuron']

    # Sort synapse by post-synaptic ID:
    def f(x, y):
        if int(x[1]) < int(y[1]):
            return -1
        elif int(x[1]) > int(y[1]):
            return 1
        else:
            return 0
    synapses.sort(cmp=f)

    s_dict = dict()
    for id, (pre_id, post_id, syn_data) in enumerate(synapses):
        syn_data = copy.deepcopy(syn_data)
        model = syn_data['model']
        syn_data['id'] = id

        if model not in s_dict:
            s_dict[model] = {k:[] for k in syn_data.keys() + ['pre', 'post']}
        
        assert set(s_dict[model].keys()) == set(syn_data.keys() + ['pre', 'post'])

        for key in syn_data.keys():
            s_dict[model][key].append(syn_data[key])
        s_dict[model]['pre'].append(pre_id)
        s_dict[model]['post'].append(post_id)

    # Return duplicate synapse model info:
    for val in s_dict.values():
        val.pop('model')
    if not s_dict:
        s_dict = dict()
        
    return n_dict, s_dict



#!/usr/bin/env python

import sys, os
from glob import glob

# Install setuptools if it isn't available:
try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from setuptools import find_packages
from setuptools import setup

NAME =               'neuroarch'
VERSION =            '0.4.2'
AUTHOR =             'Lev Givon, Nikul Ukani, Yiyin Zhou'
AUTHOR_EMAIL =       'lev@columbia.edu, nikul@ee.columbia.edu, yiyin@ee.columbia.edu'
URL =                'https://github.com/fruitflybrain/neuroarch/'
MAINTAINER =         'Yiyin Zhou'
MAINTAINER_EMAIL =   'yiyin@ee.columbia.edu'
DESCRIPTION =        'A graph-based platform for representing Drosophila brain architectures'
LONG_DESCRIPTION =   DESCRIPTION
DOWNLOAD_URL =       URL
LICENSE =            'BSD'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Database :: Front-Ends',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development']
NAMESPACE_PACKAGES = ['neuroarch']
PACKAGES =           find_packages()

docs_extras = [
    'sphinx >= 1.3',
    'sphinx_rtd_theme >= 0.1.6',
]

if __name__ == "__main__":
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    setup(
        name = NAME,
        version = VERSION,
        author = AUTHOR,
        author_email = AUTHOR_EMAIL,
        license = LICENSE,
        classifiers = CLASSIFIERS,
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        url = URL,
        maintainer = MAINTAINER,
        maintainer_email = MAINTAINER_EMAIL,
        namespace_packages = NAMESPACE_PACKAGES,
        packages = PACKAGES,
        include_package_data = True,
        install_requires = [
            'daff',
            'networkx>=2.4,<3',
            'numpy',
            'pandas',
            'path.py',
            'deepdiff',
            'tqdm'],
        extra_requires = {
            'doc': docs_extras,
        }
    )



#!/usr/bin/env python

from unittest import main, TestCase
import warnings

import neuroarch.conv.nx
import deepdiff
import pyorient
import networkx as nx

db_name = 'neuroarch_test_db'
username = 'admin'
passwd = 'admin'

class TestConvNetworkX(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = pyorient.OrientDB('localhost', 2424)
        cls.client.connect(username, passwd)
        if cls.client.db_exists(db_name):
            cls.client.db_drop(db_name)
        cls.client.db_create(db_name, pyorient.DB_TYPE_GRAPH,
                             pyorient.STORAGE_TYPE_MEMORY)
        cls.client.db_open(db_name, username, passwd)

    @classmethod
    def tearDownClass(cls):
        cls.client.connect(username, passwd)
        try:
            cls.client.db_drop(db_name)
        except Exception as e:
            warnings.warn('problem cleaning up test database: %s' % e.message)

    def setUp(self):
        cmds = ['create class neuron extends V',
                'create class synapse extends V',
                'create class data extends E']
        for cmd in cmds:
            self.client.command(cmd)

    def tearDown(self):
        cmds = ['delete vertex neuron',
                'delete vertex synapse',
                'delete edge data',
                'drop class neuron',
                'drop class synapse',
                'drop class data']
        for cmd in cmds:
            self.client.command(cmd)

    def _create_nx_graph(self):
        g = nx.MultiDiGraph()
        g.add_node(0, **{'name': 'foo',
                                 'class': 'neuron'})
        g.add_node(1, **{'name': 'bar',
                                 'class': 'neuron'})
        g.add_node(2, **{'name': 'baz',
                                 'class': 'neuron'})
        g.add_node(3, **{'name': 'foo-bar',
                                 'class': 'synapse'})
        g.add_node(4, **{'name': 'foo-baz',
                                 'class': 'synapse'})
        g.add_edge(0, 3, **{'class': 'data'})
        g.add_edge(3, 1, **{'class': 'data'})
        g.add_edge(0, 4, **{'class': 'data'})
        g.add_edge(4, 2, **{'class': 'data'})
        return g

    def _create_orient_graph(self):
        cmd = ("begin;"               
               "let foo = create vertex neuron content {'name': 'foo'};"
               "let bar = create vertex neuron content {'name': 'bar'};"
               "let baz = create vertex neuron content {'name': 'baz'};"
               "let foo_bar = create vertex synapse content {'name': 'foo-bar'};"
               "let foo_baz = create vertex synapse content {'name': 'foo-baz'};"
               "create edge data from $foo to $foo_bar;"
               "create edge data from $foo_bar to $bar;"
               "create edge data from $foo to $foo_baz;"
               "create edge data from $foo_baz to $baz;"
               "commit retry 5;")
        self.client.batch(cmd)

    def test_orient_to_nx(self):
        g_nx = self._create_nx_graph()
        self._create_orient_graph()
        g_orient = neuroarch.conv.nx.orient_to_nx(self.client,
                    'g.V.has("@class", T.in, ["neuron","synapse"])',
                    'g.E.has("@class", "data")')
        m = lambda a, b: False if deepdiff.DeepDiff(a, b) else True
        self.assertTrue(nx.isomorphism.is_isomorphic(g_nx, g_orient,
                                                     node_match=m,
                                                     edge_match=m))
        # We don't check whether the IDs of the two graphs are the same because
        # the original OrientDB graph doesn't contain any 'id' node properties.

    def test_nx_to_orient(self):
        g_nx = self._create_nx_graph()
        neuroarch.conv.nx.nx_to_orient(self.client, g_nx)
        g_orient = neuroarch.conv.nx.orient_to_nx(self.client,
                    'g.V.has("@class", T.in, ["neuron","synapse"])',
                    'g.E.has("@class", "data")')
        m = lambda a, b: False if deepdiff.DeepDiff(a, b) else True
        self.assertTrue(nx.isomorphism.is_isomorphic(g_nx, g_orient,
                                                     node_match=m,
                                                     edge_match=m))
        self.assertSetEqual(set(g_nx.nodes()), set(g_orient.nodes()))

    def test_nx_to_orient_double(self):
        g_nx = nx.MultiDiGraph()
        g_nx.add_node(0, **{'name': 'foo',
                                    'class': 'neuron',
                                    'x': 1/3.0})
        g_nx.add_node(1, **{'name': 'bar',
                                    'class': 'neuron',
                                    'x': 1/7.0})
        g_nx.add_edge(0, 1, **{'class': 'data'})
        neuroarch.conv.nx.nx_to_orient(self.client, g_nx)
        g_orient = neuroarch.conv.nx.orient_to_nx(self.client,
                    'g.V.has("@class", T.in, ["neuron","synapse"])',
                    'g.E.has("@class", "data")')
        m = lambda a, b: False if deepdiff.DeepDiff(a, b) else True
        self.assertTrue(nx.isomorphism.is_isomorphic(g_nx, g_orient,
                                                     node_match=m,
                                                     edge_match=m))
        self.assertSetEqual(set(g_nx.nodes()), set(g_orient.nodes()))
        
if __name__ == '__main__':
    main()



import pdb

PORT_IN_GPOT = 'port_in_gpot'
PORT_IN_SPK = 'port_in_spk'

def lpu_db_parser(q):
    """
    neuroArch LPU specification parser.
    
    
    Parameters
    ----------
         q : query_result

    Returns
    -------
        n_dict : dict of dict of neuron
    Each key of `n_dict` is the name of a neuron model; the values
    are dicts that map each attribute name to a list that contains the
    attribute values for each neuron.
    s_dict : dict of dict of synapse
    Each key of `s_dict` is the name of a synapse model; the values are
    dicts that map each attribute name to a list that contains the
    attribute values for each each neuron.
    
    Example
    -------
    >>> n_dict = {'LeakyIAF': {'Vr': [0.5, 0.6], 'Vt': [0.3, 0.2]},
    'MorrisLecar': {'V1': [0.15, 0.16], 'Vt': [0.13, 0.27]}}

    Notes
    -----
        All neurons must have the following attributes; any additional 
    attributes for a specific neuron model must be provided 
    for all neurons of that model type:

        1. spiking - True if the neuron emits spikes, False if it emits graded
    potentials.
    2. model - model identifier string, e.g., 'LeakyIAF', 'MorrisLecar'
    3. public - True if the neuron emits output exposed to other LPUS.
    4. extern - True if the neuron can receive external input from a file.

        All synapses must have the following attributes:

        1. class - int indicating connection class of synapse; it may assume the
    following values:

            0. spike to spike synapse
    1. spike to graded potential synapse
    2. graded potential to spike synapse
    3. graded potential to graded potential synapse
    2. model - model identifier string, e.g., 'AlphaSynapse'
    3. conductance - True if the synapse emits conductance values, False if
    it emits current values.
    4. reverse - If the `conductance` attribute is True, this attribute
    should be set to the reverse potential.

        TODO
    ----
        Input data should be validated.
    """

        
        # parse neuron data
    n_dict = {}
    neurons = [ (x._rid, x.oRecordData) for x in \
                q._client.gremlin("g.v('%s').out.has('node_type','neuron')" % (q._rid))]
    
        
        # sort based on id (id is first converted to an integer)
    # this is done so that consecutive neurons of the same type 
    # in the constructed LPU is the same in neurokernel
    neurons.sort(cmp=neuron_cmp_str)
    rid_id_map = {}
    rid_model_id_map = {}
    for id, (rid, neu) in enumerate(neurons):
        try:
            del neu['in_']
        except:
            pass
        try:
            del neu['out_']
        except:
            pass
        try:
            del neu['node_type']
        except:
            pass
        model = neu['model']

        # if an output_port, make sure selector is specified
        if 'public' not in neu.keys():
            neu['public'] = False
        if 'selector' not in neu.keys():
            neu['selector'] = ''
        # if the neuron model does not appear before, add it into n_dict
        if model not in n_dict:
            n_dict[model] = {k:[] for k in neu.keys() + ['id']}
            rid_model_id_map[model] = {}

        # neurons of the same model should have the same attributes
        assert(set(n_dict[model].keys()) == set(neu.keys() + ['id']))

        # add neuron data into the subdictionary of n_dict
        for key in neu.keys():
            n_dict[model][key].append( neu[key] )

        rid_id_map[rid] = id
        rid_model_id_map[model][rid] = len(n_dict[model]['id'])
            
        n_dict[model]['id'].append( int(id) )

    # Process output ports
    out_ports = q._client.gremlin("t = new Table();g.v('%s').out.has('node_type','port').as('x').inE.has('edge_type', 'data').outV.has('node_type','neuron').as('y').table(t).iterate();t.flatten()" % (q._rid))
            
    it = iter(out_ports)
    out_ports = [(x.oRecordData['selector'], y.oRecordData['model'], y._rid) for (x,y) in [(x,it.next()) for x in it]]

    #pdb.set_trace()
    for sel, model, rid in out_ports:
        n_dict[model]['selector'][rid_model_id_map[model][rid]] = sel

    # Process input ports
    in_ports = [(x._rid, x.oRecordData) for x in q._client.gremlin("g.v('%s').out.has('node_type','port').as('x').outE.has('edge_type', 'data').inV.has('node_type','synapse').back('x')" % (q._rid))]
            
    for id, (rid, neu) in enumerate(in_ports):
        try:
            del neu['in_']
        except:
            pass
        try:
            del neu['out_']
        except:
            pass
        try:
            del neu['node_type']
        except:
            pass
        model = neu['model']
        assert('selector' in neu.keys())

        if model == PORT_IN_GPOT:
            neu['spiking'] = False
            neu['public'] = False
        else:
            neu['spiking'] = True
            neu['public'] = False

        # if an output_port, make sure selector is specified
        if 'public' not in neu.keys():
            neu['public'] = False
        if model not in n_dict:
            n_dict[model] = {k:[] for k in neu.keys() + ['id']}
            rid_model_id_map[model] = {}

        # neurons of the same model should have the same attributes
        assert(set(n_dict[model].keys()) == set(neu.keys() + ['id']))

        # add neuron data into the subdictionary of n_dict
        for key in neu.keys():
            n_dict[model][key].append( neu[key] )

        rid_id_map[rid] = id
        rid_model_id_map[model][rid] = len(n_dict[model]['id'])
            
        n_dict[model]['id'].append( int(id) )


        
    # remove duplicate model information
    for val in n_dict.values(): val.pop('model')
    if not n_dict: n_dict = None

    synapses = q._client.gremlin("t = new Table();g.v('%s').out.has('node_type','synapse').as('x').inE.has('edge_type', 'data').outV.has('node_type','neuron').id.as('y').back('x').outE.has('edge_type','data').inV.has('node_type','neuron').id.as('z').table(t).iterate();t.flatten()" % (q._rid)) 
    it = iter(synapses)

    #pdb.set_trace()
    synapses = [ (rid_id_map[it.next().get()], rid_id_map[it.next().get()], x.oRecordData) for x in it]
        
        
    # parse synapse data
    s_dict = {}
    synapses.sort(cmp=synapse_cmp)
    scnt = 0
    for syn in synapses:
        # syn[0/1]: pre-/post-neu id; syn[2]: dict of synaptic data
        model = syn[2]['model']
        syn[2]['id'] = scnt
        # if the synapse model does not appear before, add it into s_dict
        if model not in s_dict:
            s_dict[model] = {k:[] for k in syn[2].keys() + ['pre', 'post']}

        # synapses of the same model should have the same attributes
        assert(set(s_dict[model].keys()) == set(syn[2].keys() + ['pre', 'post']))
        # add synaptic data into the subdictionary of s_dict
        for key in syn[2].keys():
            s_dict[model][key].append(syn[2][key])
        s_dict[model]['pre'].append(syn[0])
        s_dict[model]['post'].append(syn[1])
        scnt += 1
    for val in s_dict.values():
        val.pop('model')
    if not s_dict:
        s_dict = {}
    return n_dict, s_dict
                            

def neuron_cmp_str(x, y):
    if (x[0]) < (y[0]):
        return -1
    elif (x[0]) > (y[0]):
        return 1
    else:
        return 0

def synapse_cmp(x, y):
    if int(x[1]) < int(y[1]):
        return -1
    elif int(x[1]) > int(y[1]):
        return 1
    else:
        return 0



#!/usr/bin/env python

"""
Detect differences between an original and modified graph represented as Pandas
DataFrames.
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import re

import daff
import pandas as pd

def diff_edges(old, new, full_replace):
    return take_diff('edge', old, new, full_replace)
    
def diff_nodes(old, new, full_replace):
    return take_diff('node', old, new, full_replace)
    
def take_diff(diff_type, old, new, full_replace):
    """
    Find differences between original and modified DataFrame instances.

    Parameters
    ----------
    old, new : pandas.DataFrame
         DataFrame instances to compare.

    Results
    -------
    result : dict
         Dict containing keys 'add', 'mod', and 'del';
         `result['add']` contains IDs (key) and records (value) to add to `old`;
         `result['mod']` contains IDs (key) in `old` to update with
         corresponding records (value); `result['del']` contains IDs to delete
         from `old`.
    """

    assert isinstance(old, pd.DataFrame) and \
        isinstance(new, pd.DataFrame)
    
    # Move indices into first column and diff:
    if diff_type=='edge':
        old.index = old['out'].rename('id').str.cat(old['class'], sep=' ').str.cat(old['in'], sep=' ')
        new.index = new['out'].rename('id').str.cat(new['class'], sep=' ').str.cat(new['in'], sep=' ')
        new = new.drop_duplicates()
        
    # Ensure that ids in each DataFrame index are unique:
    assert len(old.index) == len(set(old.index))
    assert len(new.index) == len(set(new.index))
    
    
    # Move column names into first row:
    old = pd.concat((pd.DataFrame([old.columns], columns=old.columns), 
                   old.copy()))
    new = pd.concat((pd.DataFrame([new.columns], columns=new.columns),
                   new.copy()))

       
    data = daff.Coopy.diff(old.reset_index().values,
                           new.reset_index().values).data
    df = pd.DataFrame(data)
    # General algorithm as to how to process output of diff (`df`):
    # '->' in first col but not in second col = modify row
    # with second col as ID
    # '->' in first col and in second col = remove row
    # with ID before '->' in second col and add row with
    # ID after '->' in second col
    # '+' in first col = modify row with second col as ID
    # '+++' in first col = add row with second col as ID
    # '---' in first col = remove row with second col as ID
    # '---' or '(.*)' in the first row = modify all entries not already
    # processed

    # Entries to modify (key = ID, value = new attributes):
    mod_dict = dict()

    # Entries to delete:
    del_dict = dict()
    
    # If there is a column-wide diff operation, the @@ will appear in the second
    # row of the diff output:
    add_dict = dict()
    
    if df.ix[0, 0] == '@@':
        colname_row = 0
        start_row = 1
    else:
        edit_row = 0
        colname_row = 1
        start_row = 2
    # Entries to add:
    for i in xrange(start_row, len(df)):
        op = df.ix[i][0]
        id = df.ix[i][1]
        if op == '...' or op == '':
            continue
        elif op == '+++':
            add_dict[id] = new.ix[id].to_dict()
        elif op == '---':
            del_dict[id] = None
        elif op == '+':
            if full_replace:
                mod_dict[id] = new.ix[id].to_dict()
        elif '->' in op:
            # Renaming an ID requires deletion of the original row and creation
            # of a new row:
            if isinstance(id, str) and op in id:
                # Extract ID from 'old->new':
                old_id, new_id = re.search('(.+)%s(.+)$' % op, id).groups()

                # Preserve type of new row ID:
                old_id = old.index.dtype.type(old_id)
                new_id = new.index.dtype.type(new_id)

                del_dict[old_id] = None
                add_dict[new_id] = new.ix[new_id].to_dict()
            else:
                # If full_replace then update all fields for rows that have any change,
                # else only update the fields with changes
                if full_replace:
                    mod_dict[id] = new.ix[id].to_dict()
                else:
                    if not id in mod_dict:
                        mod_dict[id] = dict()
                    for row_val, col in zip(df.ix[i][1:], df.ix[colname_row][1:]):
                        if '->' in str(row_val):
                            mod_dict[id][col] = new.ix[id][col]
        ## cols that have --- or +++ need to have all fields removed/added
        if df.ix[0, 0] == '!':
            for col_val, col in zip(df.ix[edit_row][1:], df.ix[colname_row][1:]):
                if id not in mod_dict:
                    mod_dict[id] = dict()
                if '+++' in str(col_val):
                    mod_dict[id][col] = new.ix[id][col]
                elif '---' in str(col_val):
                    mod_dict[id][col] = None
    
    
    return {'mod': mod_dict, 'del': del_dict, 'add': add_dict}
